--[[
Copyright (c) 2023 Martin Kapfinger

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
]]--

-------------------------------------------------------------------------------
--globals
local gcodeheader_a = string.format("(Generated by the postprocessor: %s.%s)\n", ppFileBase, ppFileExt)

local gcodeheader_b = "G17 G21 G40 G49 G54\n"   --G17 XY-Plane;
                                                --G17 XY-Plane;
                                                --G21 programmed in mm (G20 inch);
                                                --G40 Tool radius compensation off;
                                                --G49 Tool length offset compensation off;
                                                --G54 Coordinate system 1;
local gcodeheader_c = "G64 G69 G80 G90 G94\n"   --G64 Default cutting mode;
                                                --G69 Turn off coordinate system rotation;
                                                --G80 Cancel canned cycle;
                                                --G90 Absolute programming;
                                                --G94 Feedrate per minute;

local g0            = "G0 "                     --Standard feed rapid
-------------------------------------------------------------------------------

local filecnt           =   0
local filelist          =   {}
local gcodeToConsole    =   false
local gcodeToFile       =   true
local filehandle        =   nil
local filepath          =   nil
local currenttool       =   nil
local emptytoolcnt      =   1

local function printcode(str)
    if gcodeToConsole then vxPrint(str) end
    if gcodeToFile   then io.write(str) end
    return nil
end

function callBegin()
    vxClear() --clear Console
    return nil
end

local function closeFilehdl()
    if filehandle ~= nil then
        io.close(filehandle)
        filehandle = nil
    end
    return nil
end

local function filefin()
    filepath = nil
    if filehandle ~= nil then
        printcode("M5\n")
        printcode("M30\n")
        closeFilehdl()
    end
    return nil
end

local function fileprepare()
    if ppDirProject      == nil or ppDirProject   == '' then return "<ppDirProject> is missing"    end
    if ppNameProject    == nil or ppNameProject == '' then return "<ppNameProject> is missing"  end
    if ppNameNcPack     == nil or ppNameNcPack  == '' then return "<ppNameNcPack> is missing"   end
    if ppNameJob        == nil or ppNameJob     == '' then return "<ppNameJob> is missing"      end
    if ppNameTool       == nil or ppNameTool    == '' then
        ppNameTool      = string.format("notool_%d", emptytoolcnt)
        emptytoolcnt    = emptytoolcnt+1    end
    if currenttool ~= ppNameTool then
        currenttool = ppNameTool
        filefin() --close current file
        fileout = string.format("%s%s_%s_%s.tap",ppDirProject, ppNameNcPack, ppNameJob, ppNameTool)
        if gcodeToFile then
            if fileout == nil or fileout == '' then
                return "no targetfile avaible" --return with failure message
            end

            filehandle = io.open(fileout,"w")   --clearing file
            if filehandle == nil then
                return(string.format("Failed to create file <%s>\nThe file may be currently being used by another program.\n", fileout))
            end
            io.close(filehandle)
            filehandle = io.open(fileout, "a")   --open file in append-mode
            if filehandle == nil then
                return(string.format("Failed to open file in append mode<%s>\n", fileout))
            end
            io.output(filehandle)
            filepath = fileout
            filelist[filecnt] = string.format("%s", filepath)
            filecnt = filecnt+1
        end
        printcode(gcodeheader_a)
        printcode(gcodeheader_b)
        printcode(gcodeheader_c)
    end
end

function callMove(x,y,z,gflag,mflag)
    if gflag ~= nil then
        if      (gflag&FLAG_G1_PLUNGE) ~= 0 then printcode(string.format("G1 F%g ",ppFeedDrill))
        elseif  (gflag&FLAG_G1_RAMP)   ~= 0 then printcode(string.format("G1 F%g ",ppFeedRamp))
        elseif  (gflag&FLAG_G1)        ~= 0 then printcode(string.format("G1 F%g ",ppFeed))
        elseif  (gflag&FLAG_G0)        ~= 0 then printcode(g0)
        end
    end
    if x ~= nil then printcode(string.format("X%.4g ",x)) end
    if y ~= nil then printcode(string.format("Y%.4g ",y)) end
    if z ~= nil then printcode(string.format("Z%.4g ",z)) end
    if mflag then
        if (mflag&FLAG_M_SPINDLE_START) ~= 0 then
            mspindlestart = "M1 (error)"
            if      (ppSpindleRotation == 0) then  mspindlestart = "M3"
            elseif  (ppSpindleRotation == 1) then  mspindlestart = "M4" end
            mcooling = ""
            if      (ppCooling == 1) then  mcooling = "M8"
            elseif  (ppCooling == 2) then  mcooling = "M7" end
            printcode(string.format("S%g %s %s ",ppSpindlespeed, mspindlestart,mcooling))
        end
    end
    printcode("\n")
    return nil
end

function callJobBegin()
    failstr = fileprepare()
    if failstr ~= nil then
        return failstr
    end
    if ppCollision == 1 then
        collisionstr =  "(!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!)\n"..
                        "(ERROR - COLLISION DETECTED)\n"..
                        "(It is strongly advised...)\n"..
                        "(...not to run this NC program!)\n"..
                        "M00\n"..
                        "M30\n"..
                        "(!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!)\n"
        if vxAsk(string.format("Job <%s>: Collision detected!\n Continue?", ppNameJob)) ~= 1 then
            return "The operation was canceled by the user."
        end
        printcode(collisionstr)
        vxPrint(collisionstr);
    end
    if ppFeedMode == 1 then g0 = string.format("G1 F%g ",ppFeedRapid)   end --optional feedmode
    --header
    if ppJobName ~= nil then
        printcode("(--------------------------------)\n")
        printcode(string.format("(job begin: %s)\n",ppJobName))
    end
    --tool change
    if ppToolId == nil then return "no tool id defined!" end
    printcode(string.format("T%d ",ppToolId))
    if ppToolName ~= nil then printcode(string.format("(%s)",ppToolName)) end
    printcode("\n")
    printcode("M6 (Toolchange)\n")
    --optional pause
    if ppPause == 1 then printcode("M1 (pause)\n") end
    --move to first position
    retval  =   callMove(nil,         nil,        ppStartposZ,FLAG_G0,nil,nil) if retval ~= nil then return retval end
    retval  =   callMove(ppStartposX, ppStartposY,nil,        FLAG_G0,nil,nil) if retval ~= nil then return retval end
    return nil
end

function callError()
    closeFilehdl() --close possibly opened file
    if filelist then
        for i=0, filecnt-1 do
          if filelist[i] == nil then vxWarn(string.format("wrong filecnt <%d>", i))
          else
            vxPrint(string.format("deleting file <%s>\n", filelist[i]))
            os.remove(filelist[i])
          end
        end
    end
    vxPrint("Operation failed.\n")
    return nil
end

function callJobEnd()
    printcode(string.format("(job end: %s)\n",ppJobName))
    printcode("(--------------------------------)\n")
    return nil
end

function callEnd()
    filefin() --close current file
    if filelist and filelist[0] then
        openfiles = vxAsk("Should the generated NC-files be opened?")
        vxPrint(string.format("Output successfully generated:\n"))
        for i=0, filecnt-1 do
          if filelist[i] == nil then vxWarn(string.format("wrong filecnt <%d>", i))
          else
            vxPrint(string.format("\t<%s>\n", filelist[i]))
            if openfiles == 1 then
                os.execute(string.format("start %s",filelist[i]))
            end
          end
        end
    end
    openexplorer = getSettingInt("pp_open_explorer")
    if openexplorer == 1 then
        ppDirProject = string.gsub(ppDirProject, "/", "\\")         --explorer need Backslash
               vxPrint(string.format("start explorer \"%s\"",ppDirProject))
        os.execute(string.format("start explorer \"%s\"",ppDirProject)) --start explorer
    end
    return nil
end

